@@ -21,7 +21,9 @@
 #include <SDL/SDL.h>
 #include <SDL/SDL_image.h>
 #include <SDL/SDL_gles.h>
+#include <SDL/SDL_ttf.h>
 #include <GLES/gl.h>
+#include <GLES/glext.h>
 
 #include <location/location-gps-device.h>
 #include <location/location-gpsd-control.h>
@@ -36,44 +38,51 @@
 #define PORTRAIT_ROTATION_Y 14
 #define LANDSCAPE_ROTATION_X 20
 
-// how much sattelite bars are drawn in status bar
+// how many sattelite bars are drawn in status bar
 #define MAX_SATTELITES 12
 
 // how many tiles horizontally
-#define TILES_X 10
+#define TILES_X 8
 
 // how many tiles vertically
-#define TILES_Y 10
+#define TILES_Y 8
 
-int tileCoordinateVisibility[TILES_X+1][TILES_Y+1];
+#define MAX_TEXTURES_CREATED_PER_FRAME 1
+
+int tileCoordinateVisibility[TILES_X + 1][TILES_Y + 1];
 
 int quit = FALSE;
 long nowMillis = 0, previousMillis = 0, diffMillis = 0;
 int downloaded = 0;
 int batteryPercentage;
-long lastBatteryRefresh;
+long lastBatteryRefresh = 0, lastGarbageCollection = 0;
 
-int syncLoadedTilesAppearance = 0;
+volatile int syncLoadedTilesAppearance = FALSE;
+int forceGarbageCollection = FALSE;
+int texturesCreatedThisFrame = 0;
 
 int currentTilesIdx = 0;
 int frame = 0;
 GLfloat fps;
 
-
 typedef enum {
     LANDSCAPE = 0, PORTRAIT = 1
-}
-    Orientation;
+} Orientation;
 
 struct {
     int mipmapToggle;
     int showGrid, showCoordinates, accelerometerEnabled;
     int showDebugTiles;
+    int locallyLoadedFadeTime, downloadedFadeTime, zoomChangeFadeTime;
+
     Orientation orientation;
     char *userAgent;
     char *referer;
 } options;
 
+TTF_Font *font = NULL;
+
+
 typedef struct {
     char* name;
     int minZoom, maxZoom;
@@ -96,14 +105,31 @@ typedef struct {
 
 struct {
     GLfloat x, y, dx, dy, scale, oldScale, destScale, rotx, roty, rotz, destRotx, destRoty, destRotz, drotz;
+    int attractedToRotZ;
     int tilex, tiley;
     int zoom, oldZoom;
     WorldCoordinate center;
+    WorldCoordinate previousCenter; // before following mode turned on
+    WorldCoordinate attraction; // position to which canvas is attracted (ie current position in following mode)
     TileProvider* provider;
     ViewMode viewMode;
     GLfloat orientationTransition;
     int orientationTransitionLinear;
     int followingMypos;
+    int attractionToPoint;
+    int rotation2d;
+    int searchBarActive;
+    int zoomBarActive;
+    long zoomBarVisibleMilis;
+    long zoomBarLastDragged;
+    GLfloat zoomBarAlpha;
+    GLfloat zoomKnotPosition;
+    GLfloat searchBarY;
+    GArray* buildings;
+    int buildingsVerticesPrecomputed;
+    GLfloat buildingsHeight;
+    GLfloat buildingsTileX;
+    GLfloat buildingsTileY;
     TileCoordinate boundingTrapezoid[4]; // points in clockwise order
 } canvas;
 
@@ -150,6 +176,13 @@ typedef struct {
     GLfloat * texCoords;
 } UiElement;
 
+struct {
+    char input[100];
+    int cursorScreen; // Cursor position on screen
+    int cursorInput; // Cursor position in input string
+    long lastChange;
+} searchBar;
+
 GQueue * downloadQueue;
 
 #include "tile.c"
@@ -158,7 +191,7 @@ GList* uiElems = NULL;
 
 UiElement* pressedUiElem = NULL;
 
-UiElement *crosshair, *zoomIn, *zoomOut, *position, *gotomypos, *view2d, *view3d;
+UiElement *crosshair, *zoomIn, *zoomOut, *position, *gotomypos, *view2d, *view3d, *compass, *search, *zoomKnot;
 
 GLfloat modelMatrix[16];
 GLfloat projectionMatrix[16];
@@ -175,6 +208,8 @@ LocationGPSDControl *control;
 #include "tileengine.c"
 #include "glcanvas.c"
 #include "accelerometer.c"
+#include "geocoder.c"
+#include "buildings3d.c"
 #include "input.c"
 #include "config.c"
 #include "location.c"
@@ -194,6 +229,11 @@ int main(int argc, char **argv) {
         exit(1);
     }
 
+    if (TTF_Init() == -1) {
+        fprintf(stderr, "TTF_Init: %s\n", TTF_GetError());
+        exit(2);
+    }
+
     surface = SDL_SetVideoMode(0, 0, 16, SDL_FULLSCREEN);
     SDL_ShowCursor(0);
 
@@ -216,6 +256,7 @@ int main(int argc, char **argv) {
         previousMillis = nowMillis;
         nowMillis = SDL_GetTicks();
         diffMillis = nowMillis - previousMillis;
+        texturesCreatedThisFrame = 0;
 
         while (SDL_PollEvent(&event)) {
             switch (event.type) {
@@ -246,33 +287,39 @@ int main(int argc, char **argv) {
         tileEngineUpdateCoordinates();
         setupGlScene();
         tileEngine_computeBoundingTrapezoid();
+        tileEngine_handleZoomHelperTiles();
         tileEngine_computeTilesVisibility();
         updateUi();
         updateSyncWaitTiles();
 
         // TODO detect screen lock
         if (!isActive) {
-            // when our window is inactive save system resources by limiting framerate to 10 fps
-            SDL_Delay(100);
+            // when our window is inactive save system resources by limiting framerate
+            SDL_Delay(50);
         }
         drawGLScene();
 
-        // refresh battery percentage every 10 seconds
+        // refresh battery percentage and save current view every 10 seconds
         if (nowMillis - lastBatteryRefresh > 10000) {
             refreshBattery();
             lastBatteryRefresh = nowMillis;
+            saveConfig();
+        }
+        if (nowMillis - lastGarbageCollection > 2000 || forceGarbageCollection || g_queue_get_length(allCreatedTiles) > 2 * TILES_X * TILES_Y) {
+            processDeferredDeallocation();
+            lastGarbageCollection = nowMillis;
+            forceGarbageCollection = FALSE;
         }
-        processDeferredDeallocation();
         frame++;
-
     }
 
     /* clean ourselves up and exit */
-    // wait for all created threads here
     saveConfig();
     SDL_GLES_DeleteContext(context);
     shutdownTileEngine();
     shutdownLocation();
+    TTF_CloseFont(font);
+    TTF_Quit();
     SDL_Quit();
     return 0;
 }
