@@ -18,9 +18,6 @@ SDL_Surface* loadTextureFromFile(char *filename, t_tile* tile) {
     SDL_Surface *surfaceRGB;
     SDL_Rect rect;
 
-//    long startTime = SDL_GetTicks();
-//    fprintf(stderr, "IMG_Load('%s')\n", filename);
-
     if ((surface = IMG_Load(filename))) {
 //        fprintf(stderr, "IMG_Load succeeded\n");
 
@@ -42,16 +39,20 @@ SDL_Surface* loadTextureFromFile(char *filename, t_tile* tile) {
         }
     } else {
         if (tile) {
+            // TODO if file exists error message should be written
+//            char* errorMessage = SDL_GetError();
+//            fprintf(stderr, "loading from file '%s' failed, file will be removed, error: %s\n", filename, errorMessage);
             remove(filename);
-//            tile -> errorMessage = SDL_GetError();
-//            fprintf(stderr, "loading from file '%s' failed in %ld ms, error: %s\n", filename, (SDL_GetTicks() - startTime), tile -> errorMessage);
         }
     }
     //free(filename);
     return NULL;
 }
 
-GLushort * convert8888to4444(SDL_Surface* surfaceRGBA) {
+double avgR = 1, avgG = 1, avgB = 1, clearR = 1, clearG = 1, clearB = 1;
+
+GLushort * convert8888to4444(SDL_Surface* surfaceRGBA, int isTile) {
+    double r=0,g=0,b=0,divisor;
     GLushort * pixels4444 = NULL;
     GLubyte* pixels8888 = surfaceRGBA->pixels;
     int i, j;
@@ -59,6 +60,29 @@ GLushort * convert8888to4444(SDL_Surface* surfaceRGBA) {
     for (i = 0; i < surfaceRGBA -> w * surfaceRGBA -> h; i++) {
         j = i << 2;
         pixels4444[i] = (pixels8888[j + 3] >> 4) + ((pixels8888[j + 2] >> 4) << 4) + ((pixels8888[j + 1] >> 4) << 8) + ((pixels8888[j] >> 4) << 12);
+        if (isTile) {
+            r += pixels8888[j ];
+            g += pixels8888[j + 1];
+            b += pixels8888[j + 2];
+        }
+        if (i % 16384 == 0) {
+            SDL_Delay(0);
+        }
+    }
+    if (isTile) {
+        r /= surfaceRGBA -> w * surfaceRGBA -> h * 256.0;
+        g /= surfaceRGBA -> w * surfaceRGBA -> h * 256.0;
+        b /= surfaceRGBA -> w * surfaceRGBA -> h * 256.0;
+        if (canvas.zoom < 1) {
+            divisor = 1;
+        } else if (canvas.zoom < 2) {
+            divisor = 4;
+        } else {
+            divisor = 8;
+        }
+        avgR = (avgR * (divisor - 1) + r) / divisor;
+        avgG = (avgG * (divisor - 1) + g) / divisor;
+        avgB = (avgB * (divisor - 1) + b) / divisor;
     }
     return pixels4444;
 }
@@ -84,11 +108,11 @@ GLuint createTexture(GLushort * pixels4444, int w, int h, int isTile) {
 
 long lastTileSync = 0;
 
-// When changing zoom level a lot of tiles are start loading at the same time. They usually
+// When changing zoom level a lot of tiles start loading at the same time. They usually
 // all load within 1 second but we want to synchronize their state change so that they will
 // start showing on screen simultaneuously.
 void updateSyncWaitTiles() {
-    if (syncLoadedTilesAppearance == 0) {
+    if (syncLoadedTilesAppearance == FALSE) {
         return;
     }
     int i, j;
@@ -119,9 +143,9 @@ void updateSyncWaitTiles() {
         }
         if (loadingCount == 0) {
             lastTileSync = 0;
+            syncLoadedTilesAppearance = FALSE;
         } else {
             lastTileSync = nowMillis;
         }
-        syncLoadedTilesAppearance = 0;
     }
 }
