@@ -20,91 +20,126 @@
 #define ALLOC_DEBUG 0
 
 typedef enum {
-    STATE_EMPTY = 0,
-    STATE_LOADING = 1,
-    STATE_ERROR = 2,
-    STATE_LOADED = 3,
-    STATE_SCALED_LOADING = 4,
-    STATE_GL_TEXTURE_NOT_CREATED = 5,
-    STATE_LOADED_SYNC_WAIT = 6
+    STATE_GARBAGE = -1, STATE_EMPTY = 0, STATE_QUEUE_DELAYED = 1, STATE_QUEUED = 2, STATE_LOADING = 3, STATE_ERROR = 4,
+    // All below are guaranteed to have pixels4444 available
+    STATE_LOADED = 5,
+    STATE_SCALED = 6,
+    STATE_GL_TEXTURE_NOT_CREATED = 7,
+    STATE_LOADED_SYNC_WAIT = 8
 } TileState;
 
 typedef struct {
     int zoom;
     int tilex, tiley;
     GLuint texture;
-    GLuint oldTexture[4];
-    GLfloat oldTextureCoords[4][8];
-    GLfloat oldTextureSize;
-    TileState state;
+    volatile TileState state;
     long stateChangeTime;
     int transitionTime;
-    GLubyte visible, deleted;
+    long textureUsedTime;
+    volatile GLubyte deleted;
+    GLubyte visible;
+    GLubyte zoomHelper;
 
-    // TODO
-    GLushort * pixels4444; // pixels used as texture, to be freed when texture is deleted
-    GLushort * oldPixels4444[4];
+    GLushort * pixels4444;
     TileProvider* provider;
-    char *errorMessage;
+    //char *errorMessage;
+    void * oldTiles[4];
 } t_tile;
 
 t_tile* tiles[TILES_X][TILES_Y][2];
 
-typedef struct {
-    int frame;
-    SDL_Surface * surface;
-    GLushort * oldPixels4444[4];
-} TDeallocationData;
+#define MAX_ZOOM_HELPER_TILES 20
 
-GQueue* deallocationQueue = NULL;
+t_tile* currentPositionZoomHelperTiles[MAX_ZOOM_HELPER_TILES];
+t_tile* destinationZoomHelperTiles[MAX_ZOOM_HELPER_TILES];
+
+GQueue* allCreatedTiles = NULL;
 
 char* getTileFileName(TileProvider *provider, t_tile *tile);
 char* getTileUrl(TileProvider *provider, t_tile *tile);
-int downloadUrl(char *url, char *filename, t_tile *tile);
-GLushort * convert8888to4444(SDL_Surface* surfaceRGBA);
+int downloadAndSave(char *url, char *filename);
+GLushort * convert8888to4444(SDL_Surface* surfaceRGBA, int isTile);
 
 gpointer dequeue(GQueue* queue);
 void deleteElems(GQueue * queue, gpointer data);
 
-void processDeferredDeallocation() {
-    TDeallocationData * data = g_queue_peek_head(deallocationQueue);
-    if (data == NULL || data -> frame + 20 > frame) {
+GQueue* garbageQueue = NULL;
+
+int isOutsideCanvas(t_tile * tile) {
+    if (tile -> zoomHelper) {
+        return FALSE;
+    }
+
+    if ((tile -> tilex < canvas.tilex) || (tile -> tiley < canvas.tiley) || (tile -> tilex >= (canvas.tilex + TILES_X)) || (tile -> tiley
+            >= (canvas.tiley + TILES_Y)) || (tile -> zoom != canvas.zoom) || (tile -> provider -> name != canvas.provider -> name)) {
+        return TRUE;
+    }
+    return FALSE;
+}
+
+void deferredDeallocation(t_tile * tile) {
+    if (tile -> state == STATE_LOADING) {
         return;
     }
-    g_queue_pop_head(deallocationQueue);
-    int i;
-    for (i = 0; i < 4; i++) {
-        if (data -> oldPixels4444[i] != NULL) {
-            if (ALLOC_DEBUG) {
-                fprintf(stderr, "deferredDeallocate: free data = %p, oldPixels4444[%d] %p\n", data, i, data -> oldPixels4444[i]);
+
+    if (nowMillis - tile -> textureUsedTime < 2000) {
+        return;
+    }
+    if (isOutsideCanvas(tile) || tile -> deleted) {
+        g_queue_remove_all(garbageQueue, tile);
+        g_queue_push_tail(garbageQueue, tile);
+    }
+}
+void deallocateTile(t_tile *tile);
+void processDeferredDeallocation() {
+    int i, j, k;
+    t_tile * tile;
+
+//    long start = SDL_GetTicks();
+    garbageQueue = g_queue_new();
+
+    // Add all outside current view to garbageQueue.
+    g_queue_foreach(allCreatedTiles, (GFunc) deferredDeallocation, NULL);
+
+    // Remove reffered by currently visible from garbageQueue.
+    for (i = 0; i < TILES_X; i++) {
+        for (j = 0; j < TILES_Y; j++) {
+            if (tiles[i][j][currentTilesIdx] != NULL && tiles[i][j][currentTilesIdx] -> visible == TRUE) {
+                tile = tiles[i][j][currentTilesIdx];
+                if (tile != NULL) {
+                    if ((tile -> state != STATE_LOADED) || ((tile -> state == STATE_LOADED) && nowMillis - tile->stateChangeTime
+                            < tile->transitionTime)) {
+                        for (k = 0; k < 4; k++) {
+                            if (tile -> oldTiles[k] != NULL) {
+                                g_queue_remove_all(garbageQueue, tile -> oldTiles[k]);
+                            }
+                        }
+                    }
+                }
             }
-            free(data -> oldPixels4444[i]);
-            data -> oldPixels4444[i] = NULL;
         }
     }
-    if (data -> surface != NULL) {
-        if (ALLOC_DEBUG) {
-            fprintf(stderr, "deferredDeallocate: free data = %p, surface %p\n", data, data -> surface);
-        }
-        SDL_FreeSurface(data -> surface);
-        data -> surface = NULL;
-    }
-    free(data);
+
+    // Deallocate all garbage.
+    //fprintf(stderr, "garbage collection (%ld ms): %d of %d tiles\n", (SDL_GetTicks() - start), g_queue_get_length(garbageQueue), g_queue_get_length(allCreatedTiles));
+    g_queue_foreach(garbageQueue, (GFunc) deallocateTile, NULL);
+    g_queue_clear(garbageQueue);
+    g_queue_free(garbageQueue);
 }
 
 void deallocateTile(t_tile *tile) {
     if (tile == NULL) {
         return;
     }
-    if (tile -> state == STATE_LOADING || tile -> state == STATE_SCALED_LOADING) {
+    if (tile -> state == STATE_LOADING) {
         tile -> deleted = TRUE;
         return;
     }
     if (ALLOC_DEBUG) {
-
         fprintf(stderr, "deallocateTile(%p)\n", tile);
     }
     deleteElems(downloadQueue, (gpointer) tile);
+    deleteElems(allCreatedTiles, (gpointer) tile);
 
     if (tile->texture != 0) {
         glDeleteTextures(1, &tile->texture);
@@ -115,49 +150,9 @@ void deallocateTile(t_tile *tile) {
         fprintf(stderr, "deallocateTile, free pixels4444 %p\n", tile -> pixels4444);
     }
 
-    // FIXME dirty hack, this should be an pointer with reference count
-    int i, j, count = 0;
-    for (i = 0; i < TILES_X; i++) {
-        for (j = 0; j < TILES_Y; j++) {
-            if (tiles[i][j][currentTilesIdx] != NULL && tiles[i][j][currentTilesIdx] != tile) {
-                if (tiles[i][j][currentTilesIdx] -> pixels4444 == tile -> pixels4444) {
-                    count++;
-                }
-            }
-        }
-    }
-
-    if (count == 0) {
-        free(tile -> pixels4444);
-    }
+    free(tile -> pixels4444);
     tile -> pixels4444 = NULL;
 
-    count = 0;
-    if (tile -> oldPixels4444[0] != NULL) {
-        // FIXME another one
-        for (i = 0; i < TILES_X; i++) {
-            for (j = 0; j < TILES_Y; j++) {
-                if (tiles[i][j][currentTilesIdx] != NULL && tiles[i][j][currentTilesIdx] != tile) {
-                    if (tiles[i][j][currentTilesIdx] -> oldPixels4444[0] == tile -> oldPixels4444[0]) {
-                        count++;
-                    }
-                }
-            }
-        }
-    }
-
-    if (count == 0) {
-        for (i = 0; i < 4; i++) {
-            if (tile -> oldPixels4444[i] != NULL) {
-                if (ALLOC_DEBUG) {
-                    fprintf(stderr, "deallocateTile, free oldpixels444[%d] %p\n", i, tile -> oldPixels4444[i]);
-                }
-                free(tile -> oldPixels4444[i]);
-                tile -> oldPixels4444[i] = NULL;
-            }
-        }
-    }
-
     //    if (tile -> errorMessage) {
     //        fprintf(stderr, "deallocateTile, free errorMessage %p\n", tile -> errorMessage);
     //        free(tile -> errorMessage);
@@ -183,15 +178,11 @@ int loadTileTexture(char* filename, t_tile* tile) {
     SDL_Surface* texture = loadTextureFromFile(filename, tile);
 
     if (texture) {
-        TDeallocationData * data;
-        data = calloc(1, sizeof(TDeallocationData));
-        data -> frame = frame;
-        data -> surface = texture;
-        tile -> pixels4444 = convert8888to4444(texture);
-        g_queue_push_tail(deallocationQueue, data);
-
+        SDL_Delay(0);
+        tile -> pixels4444 = convert8888to4444(texture, TRUE);
         //        fprintf(stderr, "loadTextureFromFile, tile= %p allocated pixels4444 %p\n", tile, tile -> pixels4444);
         tile -> state = STATE_GL_TEXTURE_NOT_CREATED;
+        SDL_FreeSurface(texture);
         return TRUE;
     }
     return FALSE;
@@ -199,16 +190,14 @@ int loadTileTexture(char* filename, t_tile* tile) {
 
 int shouldAbortLoading(t_tile * tile) {
     if (tile -> deleted) {
-        tile -> state = STATE_ERROR;
-        deallocateTile(tile);
+        tile -> state = STATE_GARBAGE;
         return TRUE;
     }
 
-    if ((tile -> tilex < canvas.tilex) || (tile -> tiley < canvas.tiley) || (tile -> tilex >= (canvas.tilex + TILES_X)) || (tile -> tiley
-            >= (canvas.tiley + TILES_Y)) || (tile -> zoom != canvas.zoom) || (tile -> provider -> name != canvas.provider -> name)) {
+    if (isOutsideCanvas(tile)) {
         //        fprintf(stderr, "tile load abort, tilex = %d, tiley = %d, canvasx = %d, canvasy = %d, tile.zoom = %d, canvas.zoom = %d\n",
         //                tile -> tilex, tile -> tiley, canvas.tilex, canvas.tiley, tile -> zoom, canvas.zoom);
-        deallocateTile(tile);
+        tile -> state = STATE_GARBAGE;
         return TRUE;
     }
     return FALSE;
@@ -244,9 +233,7 @@ int downloadThread(void* queue) {
             continue;
         }
 
-        if (tile -> state != STATE_SCALED_LOADING) {
-            tile -> state = STATE_LOADING;
-        }
+        tile -> state = STATE_LOADING;
 
         //        fprintf(stderr, "load tile start, thread id = %d\n", SDL_ThreadID());
         char *filename;
@@ -260,7 +247,7 @@ int downloadThread(void* queue) {
 
             //            fprintf(stderr, "load tile from local file failed, trying to download tile\n");
             ensureDirExists(filename);
-            if (downloadUrl(getTileUrl(canvas.provider, tile), filename, tile)) {
+            if (downloadAndSave(getTileUrl(canvas.provider, tile), filename)) {
 
                 if (shouldAbortLoading(tile)) {
                     free(filename);
@@ -268,9 +255,7 @@ int downloadThread(void* queue) {
                 }
 
                 if (loadTileTexture(filename, tile)) {
-                    if (tile ->transitionTime == 0) {
-                        tile ->transitionTime = 2000;
-                    }
+                    tile ->transitionTime = options.downloadedFadeTime;
                     //                    fprintf(stderr, "download successfull\n");
                 } else {
                     tile -> state = STATE_ERROR;
@@ -281,9 +266,7 @@ int downloadThread(void* queue) {
                 //                fprintf(stderr, "download error\n");
             }
         } else {
-            if (tile ->transitionTime == 0) {
-                tile -> transitionTime = 200;
-            }
+            tile -> transitionTime = options.locallyLoadedFadeTime;
         }
         free(filename);
     }
