@@ -11,7 +11,7 @@
  * Author: Damian Waradzyn
  */
 
-#define DOWNLOADING_THREADS 2
+#define DOWNLOADING_THREADS 4
 
 SDL_Thread* downloadThreads[DOWNLOADING_THREADS];
 
@@ -19,6 +19,8 @@ int downloadThread(void* queue);
 void tileEngine_computeBoundingTrapezoid();
 void tileEngine_computeTilesVisibility();
 
+long lastZoomChange = 0;
+
 int inline getCanvasShiftX() {
     return (TILES_X - (SCREEN_WIDTH / TILE_SIZE)) / 2;
 }
@@ -59,8 +61,8 @@ void toWorldCoordinate(TileCoordinate* tc, WorldCoordinate* result) {
 
 void canvasCenterToTileCoordinate(TileCoordinate* tc) {
     tc -> zoom = canvas.zoom;
-    tc -> tilex = canvas.tilex;
-    tc -> tiley = canvas.tiley;
+    tc -> tilex = canvas.tilex + getCanvasShiftX();
+    tc -> tiley = canvas.tiley + getCanvasShiftY();
     tc -> x = (TILE_SIZE - canvas.x) + SCREEN_WIDTH / 2;
     tc -> y = (TILE_SIZE - canvas.y) + SCREEN_HEIGHT / 2;
 }
@@ -75,14 +77,121 @@ t_tile* createTile(int i, int j) {
     tile->tiley = canvas.tiley + j;
     tile->zoom = canvas.zoom;
     tile->provider = canvas.provider;
+    tile->stateChangeTime = -1000000;
+
+    g_queue_push_tail(allCreatedTiles, tile);
     return tile;
 }
 
+long lastZoomHelperTileCreatedMilis = 0;
+void tileEngine_handleZoomHelperTiles() {
+    TileCoordinate tcMaxZoom, tc;
+    WorldCoordinate wc;
+    int i, j, newTile;
+    canvasCenterToTileCoordinate(&tcMaxZoom);
+    toWorldCoordinate(&tcMaxZoom, &wc);
+    toTileCoordinate(&wc, &tcMaxZoom, (MAX_ZOOM_HELPER_TILES - 1));
+
+    // Manage the array of zoom helper tiles - download them when applicable.
+    for (i = 0; i < MAX_ZOOM_HELPER_TILES; i++) {
+        tc.tilex = tcMaxZoom.tilex / pow(2.0, MAX_ZOOM_HELPER_TILES - i - 1);
+        tc.tiley = tcMaxZoom.tiley / pow(2.0, MAX_ZOOM_HELPER_TILES - i - 1);
+        tc.zoom = i;
+        newTile = FALSE;
+        if (currentPositionZoomHelperTiles[i] == NULL) {
+            newTile = TRUE;
+        } else {
+            if (currentPositionZoomHelperTiles[i] -> tilex != tc.tilex || currentPositionZoomHelperTiles[i] -> tiley != tc.tiley
+                    || currentPositionZoomHelperTiles[i] -> provider != canvas.provider) {
+                currentPositionZoomHelperTiles[i] -> deleted = TRUE;
+                newTile = TRUE;
+                currentPositionZoomHelperTiles[i] = NULL;
+            } else {
+                if (currentPositionZoomHelperTiles[i] -> state >= STATE_LOADED && currentPositionZoomHelperTiles[i] -> texture != 0
+                        && currentPositionZoomHelperTiles[i] -> textureUsedTime > 0 && nowMillis
+                        - currentPositionZoomHelperTiles[i] -> textureUsedTime > 3000) {
+                    //                    fprintf(stderr, "del tex %d\n", currentPositionZoomHelperTiles[i] -> texture);
+                    glDeleteTextures(1, &currentPositionZoomHelperTiles[i] -> texture);
+                    currentPositionZoomHelperTiles[i] -> texture = 0;
+                    currentPositionZoomHelperTiles[i] -> state = STATE_GL_TEXTURE_NOT_CREATED;
+                    currentPositionZoomHelperTiles[i] -> textureUsedTime = 0;
+                }
+            }
+        }
+
+        // If canvas is moving downloading higher zoom should be avoided.
+        if ((fabs(canvas.dx) > 0.2 || fabs(canvas.dy) > 0.2) && tc.zoom > canvas.zoom) {
+            newTile = FALSE;
+        }
+
+        // Limit amount of helper tiles added to queue per second.
+        if (nowMillis - lastZoomHelperTileCreatedMilis < 200) {
+            newTile = FALSE;
+        }
+
+        if (tc.zoom < canvas.provider -> minZoom || tc.zoom > canvas.provider -> maxZoom) {
+            newTile = FALSE;
+        }
+
+        if (newTile) {
+            currentPositionZoomHelperTiles[i] = createTile(0, 0);
+            currentPositionZoomHelperTiles[i] -> state = STATE_QUEUED;
+            currentPositionZoomHelperTiles[i] -> visible = FALSE;
+            currentPositionZoomHelperTiles[i] -> zoomHelper = TRUE;
+            currentPositionZoomHelperTiles[i] -> tilex = tc.tilex;
+            currentPositionZoomHelperTiles[i] -> tiley = tc.tiley;
+            currentPositionZoomHelperTiles[i] -> zoom = tc.zoom;
+            enqueue(downloadQueue, currentPositionZoomHelperTiles[i]);
+            lastZoomHelperTileCreatedMilis = nowMillis;
+        }
+    }
+
+    // Go through all currently displayed tiles and set oldtile where applicable.
+    for (i = 0; i < TILES_X; i++) {
+        for (j = 0; j < TILES_Y; j++) {
+            if (tiles[i][j][currentTilesIdx] != NULL) {
+                if (tiles[i][j][currentTilesIdx] -> texture == 0 && tiles[i][j][currentTilesIdx] -> oldTiles[0] == NULL) {
+                    int k, l = 1;
+                    for (k = (tiles[i][j][currentTilesIdx] -> zoom); k >= 0; k--, l *= 2) {
+                        if (k < MAX_ZOOM_HELPER_TILES) {
+                            if (currentPositionZoomHelperTiles[k] != NULL) {
+                                if (tiles[i][j][currentTilesIdx] -> tilex >= 0 && tiles[i][j][currentTilesIdx] -> tiley >= 0
+                                        && (tiles[i][j][currentTilesIdx] -> tilex / l == currentPositionZoomHelperTiles[k] -> tilex)
+                                        && (tiles[i][j][currentTilesIdx] -> tiley / l == currentPositionZoomHelperTiles[k] -> tiley)) {
+                                    if (currentPositionZoomHelperTiles[k] -> state >= STATE_LOADED) {
+                                        if (currentPositionZoomHelperTiles[k] -> texture == 0) {
+                                            if (texturesCreatedThisFrame >= MAX_TEXTURES_CREATED_PER_FRAME) {
+                                                continue;
+                                            }
+                                            currentPositionZoomHelperTiles[k] -> texture = createTexture(
+                                                    currentPositionZoomHelperTiles[k] -> pixels4444, TILE_SIZE, TILE_SIZE, TRUE);
+                                            currentPositionZoomHelperTiles[k] -> state = STATE_LOADED;
+                                            texturesCreatedThisFrame++;
+                                        }
+                                    }
+                                    tiles[i][j][currentTilesIdx] -> oldTiles[0] = currentPositionZoomHelperTiles[k];
+                                    break;
+                                }
+                            }
+                        }
+                    }
+                }
+
+                // If the queueing was delayed enqueue now.
+                if (tiles[i][j][currentTilesIdx] -> state == STATE_QUEUE_DELAYED && nowMillis - lastZoomChange > 1500) {
+                    tiles[i][j][currentTilesIdx] -> state = STATE_QUEUED;
+                    enqueue(downloadQueue, tiles[i][j][currentTilesIdx]);
+                }
+            }
+        }
+    }
+}
+
 void initTileEngine() {
     int i, j;
 
     downloadQueue = createQueue();
-    deallocationQueue = g_queue_new();
+    allCreatedTiles = g_queue_new();
 
     canvas.scale = 1.0;
     canvas.viewMode = VIEW_2D;
@@ -106,6 +215,7 @@ void initTileEngine() {
             enqueue(downloadQueue, tiles[i][j][currentTilesIdx]);
         }
     }
+    tileEngine_handleZoomHelperTiles();
 }
 
 int inline roundDown(double x) {
@@ -129,21 +239,26 @@ int adjustShifts() {
 
 void recreateTiles() {
     purgeQueue(downloadQueue);
-    int i, j;
+    int i, j, k;
+
+    // start loading new tile textures:
+    //    syncLoadedTilesAppearance = TRUE;
     for (i = 0; i < TILES_X; i++) {
         for (j = 0; j < TILES_Y; j++) {
+            tiles[i][j][1 - currentTilesIdx] = createTile(i, j);
             if (tiles[i][j][currentTilesIdx] != NULL) {
-                deallocateTile(tiles[i][j][currentTilesIdx]);
-                tiles[i][j][currentTilesIdx] = NULL;
+                if (tiles[i][j][currentTilesIdx] -> texture != 0) {
+                    tiles[i][j][1 - currentTilesIdx] -> oldTiles[0] = tiles[i][j][currentTilesIdx];
+                    tiles[i][j][1 - currentTilesIdx] -> stateChangeTime = nowMillis;
+                } else {
+                    for (k = 0; k < 4; k++) {
+                        tiles[i][j][1 - currentTilesIdx] -> oldTiles[k] = tiles[i][j][currentTilesIdx]-> oldTiles[k];
+                    }
+                }
             }
         }
     }
-    // start loading new tile textures:
-    for (i = 0; i < TILES_X; i++) {
-        for (j = 0; j < TILES_Y; j++) {
-            tiles[i][j][currentTilesIdx] = createTile(i, j);
-        }
-    }
+    currentTilesIdx = 1 - currentTilesIdx;
     tileEngine_computeBoundingTrapezoid();
     tileEngine_computeTilesVisibility();
     for (i = 0; i < TILES_X; i++) {
@@ -161,18 +276,10 @@ void handleZoomChange() {
 
     toTileCoordinate(&canvas.center, &tc, canvas.zoom);
 
-    if (canvas.zoom - canvas.oldZoom == 1) {
-        canvas.tilex = tc.tilex + getCanvasShiftX();
-        canvas.tiley = tc.tiley + getCanvasShiftY();
-        canvas.x = tc.x - SCREEN_WIDTH / 2;
-        canvas.y = tc.y - SCREEN_HEIGHT / 2;
-    } else {
-        canvas.tilex = tc.tilex - getCanvasShiftX();
-        canvas.tiley = tc.tiley - getCanvasShiftY();
-        // TODO - where do those number (-16 +32) come from?
-        canvas.x = tc.x - 16;
-        canvas.y = tc.y + SCREEN_HEIGHT / 2 + 32;
-    }
+    canvas.tilex = tc.tilex - getCanvasShiftX();
+    canvas.tiley = tc.tiley - getCanvasShiftY();
+    canvas.x = tc.x - SCREEN_WIDTH / 2;
+    canvas.y = tc.y - SCREEN_HEIGHT / 2;
 
     adjustShifts();
 
@@ -183,7 +290,8 @@ void handleZoomChange() {
         canvas.y = TILE_SIZE - canvas.y;
     }
 
-    syncLoadedTilesAppearance = TRUE;
+    //    syncLoadedTilesAppearance = TRUE;
+    lastZoomChange = nowMillis;
 
     if (canvas.zoom - canvas.oldZoom == 1) {
         //        fprintf(stderr, "zoomIn 1\n");
@@ -199,83 +307,62 @@ void handleZoomChange() {
 
                 if (oldi >= 0 && oldj >= 0) {
 
-//                    if (tiles[oldi / 2][oldj / 2][currentTilesIdx] != NULL && tiles[oldi / 2][oldj / 2][currentTilesIdx] -> state == STATE_LOADED) {
-//                        tiles[i][j][1 - currentTilesIdx] -> state = STATE_SCALED_LOADING;
-//                        tiles[i][j][1 - currentTilesIdx] -> oldTextureSize = TILE_SIZE;
-//
-//                        tiles[i][j][1 - currentTilesIdx] -> oldTexture[0] = tiles[oldi / 2][oldj / 2][currentTilesIdx] -> texture;
-//                        tiles[i][j][1 - currentTilesIdx] -> oldPixels4444[0] = tiles[oldi / 2][oldj / 2][currentTilesIdx] ->pixels4444;
-//
-//                        tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[0][0] = (oldi % 2) * 0.5;
-//                        tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[0][1] = (oldj % 2) * 0.5;
-//
-//                        tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[0][2] = ((oldi % 2) + 1) * 0.5;
-//                        tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[0][3] = (oldj % 2) * 0.5;
-//
-//                        tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[0][4] = (oldi % 2) * 0.5;
-//                        tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[0][5] = ((oldj % 2) + 1) * 0.5;
-//
-//                        tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[0][6] = ((oldi % 2) + 1) * 0.5;
-//                        tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[0][7] = ((oldj % 2) + 1) * 0.5;
-//                        tiles[i][j][1 - currentTilesIdx] -> transitionTime = 1500;
-//                        enqueue(downloadQueue, tiles[i][j][1 - currentTilesIdx]);
-//                    } else {
-                        deallocateTile(tiles[oldi / 2][oldj / 2][currentTilesIdx]);
+                    if (tiles[oldi / 2][oldj / 2][currentTilesIdx] != NULL && tiles[oldi / 2][oldj / 2][currentTilesIdx] -> state >= STATE_LOADED) {
+                        tiles[i][j][1 - currentTilesIdx] -> oldTiles[0] = tiles[oldi / 2][oldj / 2][currentTilesIdx];
+
+                        if (tiles[oldi / 2][oldj / 2][currentTilesIdx] -> state == STATE_GL_TEXTURE_NOT_CREATED) {
+                            tiles[oldi / 2][oldj / 2][currentTilesIdx] -> texture = createTexture(
+                                    tiles[oldi / 2][oldj / 2][currentTilesIdx] -> pixels4444, TILE_SIZE, TILE_SIZE, TRUE);
+                        }
+
+                        tiles[oldi / 2][oldj / 2][currentTilesIdx] -> state = STATE_SCALED;
+                        tiles[i][j][1 - currentTilesIdx] -> transitionTime = options.zoomChangeFadeTime;
+                        if (canvas.zoomBarLastDragged - nowMillis < 500) {
+                            tiles[i][j][1 - currentTilesIdx] -> state = STATE_QUEUE_DELAYED;
+                        } else {
+                            tiles[i][j][1 - currentTilesIdx] -> state = STATE_QUEUED;
+                            enqueue(downloadQueue, tiles[i][j][1 - currentTilesIdx]);
+                        }
+                    } else {
                         tiles[oldi / 2][oldj / 2][currentTilesIdx] = NULL;
-                        tiles[i][j][1 - currentTilesIdx] -> transitionTime = 1500;
-                        enqueue(downloadQueue, tiles[i][j][1 - currentTilesIdx]);
-//                    }
-                }
-            }
-        }
-        currentTilesIdx = 1 - currentTilesIdx;
-    } else if (canvas.zoom - canvas.oldZoom == -1) {
-        int oldtilex = tiles[0][0][currentTilesIdx] -> tilex;
-        int oldtiley = tiles[0][0][currentTilesIdx] -> tiley;
-        int ii, jj;
-
-        for (i = 0; i < TILES_X; i++) {
-            for (j = 0; j < TILES_Y; j++) {
-                tiles[i][j][1 - currentTilesIdx] = createTile(i, j);
-
-                for (ii = 0; ii < 2; ii++) {
-                    for (jj = 0; jj < 2; jj++) {
-                        int oldi = (tiles[i][j][1 - currentTilesIdx] -> tilex - (oldtilex / 2)) * 2 + ii - (oldtilex % 2);
-                        int oldj = (tiles[i][j][1 - currentTilesIdx] -> tiley - (oldtiley / 2)) * 2 + jj - (oldtiley % 2);
-
-                        if (oldi >= 0 && oldi < TILES_X && oldj >= 0 && oldj < TILES_Y) {
-                            if (tiles[oldi][oldj][currentTilesIdx] != NULL) {
-                                if (tiles[oldi][oldj][currentTilesIdx] -> state == STATE_LOADED) {
-                                    tiles[i][j][1 - currentTilesIdx] -> oldTextureSize = TILE_SIZE / 2;
-                                    tiles[i][j][1 - currentTilesIdx] -> state = STATE_SCALED_LOADING;
-
-                                    tiles[i][j][1 - currentTilesIdx] -> oldTexture[ii * 2 + jj] = tiles[oldi][oldj][currentTilesIdx] -> texture;
-                                    tiles[i][j][1 - currentTilesIdx] -> oldPixels4444[ii * 2 + jj] = tiles[oldi][oldj][currentTilesIdx] -> pixels4444;
-
-                                    tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[ii * 2 + jj][0] = 0.0;
-                                    tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[ii * 2 + jj][1] = 0.0;
-
-                                    tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[ii * 2 + jj][2] = 1.0;
-                                    tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[ii * 2 + jj][3] = 0.0;
-
-                                    tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[ii * 2 + jj][4] = 0.0;
-                                    tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[ii * 2 + jj][5] = 1.0;
-
-                                    tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[ii * 2 + jj][6] = 1.0;
-                                    tiles[i][j][1 - currentTilesIdx] -> oldTextureCoords[ii * 2 + jj][7] = 1.0;
-                                }
-                            }
+                        tiles[i][j][1 - currentTilesIdx] -> transitionTime = options.zoomChangeFadeTime;
+                        if (canvas.zoomBarLastDragged - nowMillis < 500) {
+                            tiles[i][j][1 - currentTilesIdx] -> state = STATE_QUEUE_DELAYED;
+                        } else {
+                            tiles[i][j][1 - currentTilesIdx] -> state = STATE_QUEUED;
+                            enqueue(downloadQueue, tiles[i][j][1 - currentTilesIdx]);
                         }
                     }
                 }
-                tiles[i][j][1 - currentTilesIdx] -> transitionTime = 1500;
-                enqueue(downloadQueue, tiles[i][j][1 - currentTilesIdx]);
             }
         }
         currentTilesIdx = 1 - currentTilesIdx;
+        if (canvas.destScale != 0.0) {
+            canvas.destScale /= 2.0;
+        }
+    } else if (canvas.zoom - canvas.oldZoom == -1) {
+        for (i = 0; i < TILES_X; i++) {
+            for (j = 0; j < TILES_Y; j++) {
+                tiles[i][j][1 - currentTilesIdx] = createTile(i, j);
+                if (canvas.zoomBarLastDragged - nowMillis < 500) {
+                    tiles[i][j][1 - currentTilesIdx] -> state = STATE_QUEUE_DELAYED;
+                } else {
+                    tiles[i][j][1 - currentTilesIdx] -> state = STATE_QUEUED;
+                    enqueue(downloadQueue, tiles[i][j][1 - currentTilesIdx]);
+                }
+                tiles[i][j][1 - currentTilesIdx] -> transitionTime = options.zoomChangeFadeTime;
+            }
+        }
+        currentTilesIdx = 1 - currentTilesIdx;
+        if (canvas.destScale != 0.0) {
+            canvas.destScale *= 2.0;
+        }
     } else {
+        // Normally not used because changes to zoom level are supported only by one level at a time.
+        // Useful for changing between tile providers that have have no zoom level common.
         recreateTiles();
     }
+    forceGarbageCollection = TRUE;
 }
 
 int inline getLowerDeallocateBound(int size, int shift) {
@@ -291,13 +378,13 @@ void deallocateTiles(int horizontalShift, int verticalShift) {
     int i, j;
     for (i = getLowerDeallocateBound(TILES_X, horizontalShift); i < getUpperDeallocateBound(TILES_X, horizontalShift); i++) {
         for (j = 0; j < TILES_Y; j++) {
-            deallocateTile(tiles[i][j][currentTilesIdx]);
+            //            deallocateTile(tiles[i][j][currentTilesIdx]);
             tiles[i][j][currentTilesIdx] = NULL;
         }
     }
     for (i = 0; i < TILES_X; i++) {
         for (j = getLowerDeallocateBound(TILES_Y, verticalShift); j < getUpperDeallocateBound(TILES_Y, verticalShift); j++) {
-            deallocateTile(tiles[i][j][currentTilesIdx]);
+            //            deallocateTile(tiles[i][j][currentTilesIdx]);
             tiles[i][j][currentTilesIdx] = NULL;
         }
     }
@@ -328,10 +415,14 @@ void tileEngineProcessMouse(int uiElemPressed) {
 
     if (uiElemPressed == FALSE) {
         if (mouse.button) {
-            canvas.followingMypos = FALSE;
+            canvas.followingMypos = 0;
+            canvas.attractionToPoint = 0;
+            if (canvas.viewMode == VIEW_3D || canvas.rotation2d == 1) {
+                canvas.attractedToRotZ = 0;
+            }
         }
 
-        if (canvas.viewMode == VIEW_3D) {
+        if (canvas.viewMode == VIEW_3D || canvas.rotation2d == 1) {
             if (options.orientation == PORTRAIT && mouse.x < 150) {
                 canvas.drotz -= mouse.ydelta / 18.0;
                 return;
@@ -366,17 +457,21 @@ void tileEngineProcessAccelerometer() {
         if (canvas.viewMode == VIEW_2D) {
             if (options.orientation == LANDSCAPE) {
                 if (abs(accelerometer.x) > 70) {
-                    canvas.dx += accelerometer.x / 1000.0;
+                    canvas.dx += accelerometer.x / 1000.0 * sin((canvas.rotz + 90) * M_PI / 180.0);
+                    canvas.dy += accelerometer.x / 1000.0 * cos((canvas.rotz + 90) * M_PI / 180.0);
                 }
                 if (abs(accelerometer.y) > 70) {
-                    canvas.dy += accelerometer.y / 1000.0;
+                    canvas.dx += accelerometer.y / 1000.0 * sin(canvas.rotz * M_PI / 180.0);
+                    canvas.dy += accelerometer.y / 1000.0 * cos(canvas.rotz * M_PI / 180.0);
                 }
             } else {
                 if (abs(accelerometer.y) > 70) {
-                    canvas.dx -= accelerometer.y / 1000.0;
+                    canvas.dx += accelerometer.y / 1000.0 * sin(canvas.rotz * M_PI / 180.0);
+                    canvas.dy += accelerometer.y / 1000.0 * cos(canvas.rotz * M_PI / 180.0);
                 }
                 if (abs(accelerometer.x) > 70) {
-                    canvas.dy += accelerometer.x / 1000.0;
+                    canvas.dx += accelerometer.x / 1000.0 * sin((canvas.rotz + 90) * M_PI / 180.0);
+                    canvas.dy += accelerometer.x / 1000.0 * cos((canvas.rotz + 90) * M_PI / 180.0);
                 }
             }
         } else {
@@ -393,8 +488,14 @@ void tileEngineProcessAccelerometer() {
             }
             if (options.orientation == LANDSCAPE) {
                 canvas.drotz += accelerometer.x / 2000.0;
+                if (abs(accelerometer.x) > 70) {
+                    canvas.attractedToRotZ = 0;
+                }
             } else {
                 canvas.drotz -= accelerometer.y / 2000.0;
+                if (abs(accelerometer.y) > 70) {
+                    canvas.attractedToRotZ = 0;
+                }
             }
         }
     }
@@ -406,6 +507,56 @@ void tileEngineProcessAccelerometer() {
     }
 }
 
+void updateCanvasCenterWorldCoordinate() {
+    TileCoordinate tc;
+    canvasCenterToTileCoordinate(&tc);
+
+    int horizontalShift = roundDown(tc.x / (double) TILE_SIZE) - getCanvasShiftX();
+    int verticalShift = roundDown(tc.y / (double) TILE_SIZE) - getCanvasShiftY();
+
+    tc.x -= horizontalShift * TILE_SIZE;
+    tc.y -= verticalShift * TILE_SIZE;
+
+    tc.tilex += horizontalShift;
+    tc.tiley += verticalShift;
+
+    toWorldCoordinate(&tc, &canvas.center);
+}
+
+void tileEngineZoomKnotPressed() {
+    canvas.zoomBarLastDragged = nowMillis;
+    canvas.zoomBarVisibleMilis = nowMillis + 10000;
+}
+
+// Calculates zoom level from zoom knot position.
+int getKnotDestZoom() {
+    GLfloat screenSize = canvas.orientationTransitionLinear < 80 ? SCREEN_WIDTH : SCREEN_HEIGHT;
+    int zoomLevelCount = canvas.provider -> maxZoom - canvas.provider -> minZoom + 1;
+    return (canvas.orientationTransitionLinear < 80 ? zoomKnot -> x - 82 : (SCREEN_HEIGHT - zoomKnot -> y - 82 - 64)) / ((screenSize - 232)
+            / zoomLevelCount) + canvas.provider -> minZoom;
+}
+
+void tileEngineZoomKnotDragged() {
+    canvas.zoomBarActive = TRUE;
+    canvas.zoomBarLastDragged = nowMillis;
+    canvas.zoomBarVisibleMilis = nowMillis + 7000;
+    canvas.zoomKnotPosition += canvas.orientationTransitionLinear < 80 ? mouse.x - mouse.oldx : mouse.y - mouse.oldy;
+    if (canvas.zoomKnotPosition < 100 - 18) {
+        canvas.zoomKnotPosition = 100 - 18;
+    }
+    int screenSize = canvas.orientationTransitionLinear < 80 ? SCREEN_WIDTH : SCREEN_HEIGHT;
+    if (canvas.zoomKnotPosition > screenSize - 100 - 32 - 14) {
+        canvas.zoomKnotPosition = screenSize - 100 - 32 - 14;
+    }
+    if (canvas.zoom != getKnotDestZoom()) {
+        canvas.destScale = pow(2.0, getKnotDestZoom() - canvas.zoom);
+    }
+}
+
+void tileEngineToggleSearchBar() {
+    canvas.searchBarActive = 1 - canvas.searchBarActive;
+}
+
 void tileEngineUpdateCoordinates() {
     //int zoomChange = (int) round(log2((double) canvas.scale));
     int zoomChange = 0;
@@ -415,40 +566,64 @@ void tileEngineUpdateCoordinates() {
 
     if (canvas.followingMypos) {
         if (device -> fix -> fields | LOCATION_GPS_DEVICE_LATLONG_SET) {
-            GLfloat x, y;
-            WorldCoordinate wc;
-            wc.latitude = device -> fix -> latitude;
-            wc.longitude = device -> fix -> longitude;
-            TileCoordinate tc;
-            toTileCoordinate(&wc, &tc, canvas.zoom);
-            toScreenCoordinate(&tc, &x, &y);
-            //            fprintf(stderr, "x = %f, y = %f\t\ttilex = %d, tiley = %d, tc.x = %f, tc.y = %f \n", x, y, tc.tilex, tc.tiley, tc.x, tc.y);
-            canvas.dx = (SCREEN_WIDTH / 2.0 - (canvas.x - TILE_SIZE + x)) / 20.0;
-            canvas.dy = (SCREEN_HEIGHT / 2.0 - (canvas.y - TILE_SIZE + y)) / 20.0;
-            if (canvas.viewMode == VIEW_3D) {
-                if ((device -> fix -> fields | LOCATION_GPS_DEVICE_TRACK_SET) && device -> satellites_in_use > 0 && device -> fix -> speed > 7) {
-                    canvas.rotz += (device -> fix -> track - canvas.rotz) * 0.03;
+            canvas.attraction.latitude = device -> fix -> latitude;
+            canvas.attraction.longitude = device -> fix -> longitude;
+            canvas.attractionToPoint = 1;
+        }
+        if (canvas.viewMode == VIEW_3D || canvas.rotation2d == 1) {
+            if ((device -> fix -> fields | LOCATION_GPS_DEVICE_TRACK_SET) && device -> satellites_in_use > 0 && device -> fix -> speed > 3) {
+                if (options.orientation == LANDSCAPE) {
+                    canvas.destRotz = -device -> fix -> track;
+                } else {
+                    canvas.destRotz = 270 - device -> fix -> track;
                 }
+                canvas.attractedToRotZ = 1;
             }
         }
     }
 
-    if (canvas.destScale != 0.0) {
-        GLfloat tmp = canvas.destScale - canvas.scale;
-        canvas.oldScale = canvas.scale;
-        canvas.scale += tmp / 5;
+    if (canvas.attractionToPoint) {
+        GLfloat x, y;
+        TileCoordinate tc;
+        toTileCoordinate(&canvas.attraction, &tc, canvas.zoom);
+        toScreenCoordinate(&tc, &x, &y);
+        //            fprintf(stderr, "x = %f, y = %f\t\ttilex = %d, tiley = %d, tc.x = %f, tc.y = %f \n", x, y, tc.tilex, tc.tiley, tc.x, tc.y);
+        canvas.dx = (SCREEN_WIDTH / 2.0 - (canvas.x - TILE_SIZE + x)) / 20.0;
+        canvas.dy = (SCREEN_HEIGHT / 2.0 - (canvas.y - TILE_SIZE + y)) / 20.0;
+    }
 
-        if (fabs(tmp) < 0.001) {
+    if (canvas.destScale != 0.0) {
+        GLfloat slowingDownSpeedFactor = 10.0;
+        GLfloat tmp = canvas.destScale / canvas.scale;
+        canvas.oldScale = canvas.scale;
+
+        //        GLfloat maxSpeedBorder = 1.0 + ((canvas.maxScaleSpeed - 1.0) * slowingDownSpeedFactor);
+        //        if (tmp > maxSpeedBorder) {
+        //            canvas.scale *= canvas.maxScaleSpeed;
+        //} else if (tmp < 1.0 / maxSpeedBorder || tmp < 0.) {
+        //    canvas.scale /= canvas.maxScaleSpeed;
+        //        } else {
+        if (tmp < -slowingDownSpeedFactor + 1.0) {
+            canvas.scale = canvas.destScale;
+            canvas.destScale = 0.0;
+        } else {
+            canvas.scale *= 1.0 + (tmp - 1.0) / slowingDownSpeedFactor;
+        }
+        //        }
+
+        if (fabs(tmp - 1.0) < 0.01 && (fabs(canvas.destScale - 1.0) < 0.01 || fabs(canvas.destScale - 2.0) < 0.01 || fabs(canvas.destScale - .5)
+                < 0.01)) {
             canvas.scale = canvas.destScale;
             canvas.destScale = 0.0;
         }
-    } else {
-        if (canvas.scale < 1.0) {
-            zoomChange = -1;
-        }
-        if (canvas.scale >= 2.0) {
-            zoomChange = 1;
-        }
+        //fprintf(stderr, "%d\t%f\t%f\n", canvas.zoom, canvas.oldScale, canvas.scale);
+    }
+    if (canvas.scale < 1.0) {
+        zoomChange = -1;
+    }
+
+    if (canvas.scale >= 2.0) {
+        zoomChange = 1;
     }
 
     if ((canvas.zoom + zoomChange) < canvas.provider -> minZoom || (canvas.zoom + zoomChange) > canvas.provider -> maxZoom) {
@@ -469,25 +644,23 @@ void tileEngineUpdateCoordinates() {
         }
     }
 
-    TileCoordinate tc;
-    canvasCenterToTileCoordinate(&tc);
-
-    int horizontalShift = roundDown(tc.x / (double) TILE_SIZE);
-    int verticalShift = roundDown(tc.y / (double) TILE_SIZE);
-
-    tc.x -= horizontalShift * TILE_SIZE;
-    tc.y -= verticalShift * TILE_SIZE;
-
-    tc.tilex += horizontalShift;
-    tc.tiley += verticalShift;
-
-    toWorldCoordinate(&tc, &canvas.center);
+    updateCanvasCenterWorldCoordinate();
 
     if (zoomChange != 0 /* && nowMillis - lastZoomChangeMilis > 500 */) {
         canvas.oldZoom = canvas.zoom;
         canvas.zoom += zoomChange;
         handleZoomChange();
-        canvas.scale /= pow(2.0, zoomChange);
+        //        if (zoomChange > 0) {
+        //            canvas.scale /= 4.0/3.0;
+        //        } else {
+        //            canvas.scale *= 4.0/3.0;
+        //        }
+        //        canvas.scale /= pow(2.0, zoomChange);
+        if (zoomChange == 1) {
+            canvas.scale /= 2.0;
+        } else if (zoomChange == -1) {
+            canvas.scale *= 2.0;
+        }
     }
 
     canvas.rotx += (canvas.destRotx - canvas.rotx) * 0.03;
@@ -502,17 +675,27 @@ void tileEngineUpdateCoordinates() {
             canvas.orientationTransitionLinear++;
         }
     }
-    if (canvas.viewMode == VIEW_2D) {
-        canvas.rotz += (canvas.destRotz - canvas.rotz) * 0.03;
+    if (canvas.attractedToRotZ) {
+        double dist = canvas.destRotz - canvas.rotz, dist2 = 180.0;
+        if (canvas.rotz < 0.0 && canvas.destRotz > 0.0) {
+            dist2 = -(180.0 + canvas.rotz) - (180.0 - canvas.destRotz);
+        } else if (canvas.rotz > 0.0 && canvas.destRotz < 0.0) {
+            dist2 = 180 - canvas.rotz + (180.0 + canvas.destRotz);
+        }
+        if (fabs(dist) < fabs(dist2)) {
+            canvas.rotz += dist * 0.03;
+        } else {
+            canvas.rotz += dist2 * 0.03;
+        }
     } else {
         canvas.rotz += canvas.drotz;
         canvas.drotz *= 0.9;
-        if (canvas.rotz > 180) {
-            canvas.rotz -= 360;
-        }
-        if (canvas.rotz < -180) {
-            canvas.rotz += 360;
-        }
+    }
+    if (canvas.rotz > 180) {
+        canvas.rotz -= 360;
+    }
+    if (canvas.rotz < -180) {
+        canvas.rotz += 360;
     }
     if (canvas.orientationTransition > 0.999) {
         canvas.orientationTransition = 1.0;
@@ -521,32 +704,67 @@ void tileEngineUpdateCoordinates() {
     }
 }
 
+void activateZoomBar() {
+    canvas.zoomBarActive = TRUE;
+    canvas.zoomBarVisibleMilis = nowMillis + 3000;
+    int zoomLevelCount = canvas.provider -> maxZoom - canvas.provider -> minZoom + 2;
+    GLfloat knotZoomLevel = canvas.zoom - canvas.provider -> minZoom;
+    if (canvas.destScale != 0.0) {
+        knotZoomLevel += round(log2(canvas.destScale));
+    }
+    if (canvas.orientationTransitionLinear < 80) {
+        canvas.zoomKnotPosition = knotZoomLevel * (SCREEN_WIDTH - 228) / zoomLevelCount + 114;
+    } else {
+        canvas.zoomKnotPosition = -(knotZoomLevel * (SCREEN_HEIGHT - 232) / zoomLevelCount) + SCREEN_HEIGHT - 114-32-14;
+    }
+}
+
+long zoomInLastMilis = 0;
 void tileEngineZoomIn() {
     if (SDL_GetModState() & KMOD_SHIFT) {
         canvas.oldScale = canvas.scale;
         canvas.scale *= 1.005;
     } else {
-        if (canvas.destScale == 0.0) {
-            if (canvas.scale <= 0.5) {
-                canvas.destScale = 1.0;
-            } else {
-                canvas.destScale = 2.0;
+        if (nowMillis - zoomInLastMilis > 100) {
+            if (canvas.destScale == 0.0) {
+                if (canvas.scale <= 0.5) {
+                    canvas.destScale = 1.0;
+                } else {
+                    canvas.destScale = 2.0;
+                }
+            } else if (canvas.scale < 2.0 && canvas.destScale <= 2.0 && canvas.zoom < canvas.provider -> maxZoom) {
+                canvas.destScale *= 2.0;
             }
+            zoomInLastMilis = nowMillis;
+            if (mouse.oldButton == 0 && pressedUiElem == zoomIn) {
+                zoomInLastMilis += 400;
+            }
+            activateZoomBar();
         }
     }
 }
 
+long zoomOutLastMilis = 0;
 void tileEngineZoomOut() {
     if (SDL_GetModState() & KMOD_SHIFT) {
         canvas.oldScale = canvas.scale;
         canvas.scale /= 1.005;
     } else {
-        if (canvas.destScale == 0.0) {
-            if (canvas.scale >= 2.0) {
-                canvas.destScale = 1.0;
-            } else {
-                canvas.destScale = 0.5;
+        if (nowMillis - zoomOutLastMilis > 100) {
+            if (canvas.destScale == 0.0) {
+                if (canvas.scale >= 2.0) {
+                    canvas.destScale = 1.0;
+                } else {
+                    canvas.destScale = 0.5;
+                }
+            } else if (canvas.scale > 0.25 && canvas.destScale >= 0.25 && canvas.zoom > canvas.provider -> minZoom) {
+                canvas.destScale /= 2.0;
             }
+            zoomOutLastMilis = nowMillis;
+            if (mouse.oldButton == 0 && pressedUiElem == zoomOut) {
+                zoomOutLastMilis += 400;
+            }
+            activateZoomBar();
         }
     }
 }
@@ -554,6 +772,11 @@ void tileEngineZoomOut() {
 void tileEngineGotomypos() {
     if (mouse.oldButton == 0) {
         canvas.followingMypos = 1 - canvas.followingMypos;
+        if (canvas.followingMypos) {
+            updateCanvasCenterWorldCoordinate();
+            canvas.previousCenter.latitude = canvas.center.latitude;
+            canvas.previousCenter.longitude = canvas.center.longitude;
+        }
     }
 }
 
@@ -564,11 +787,17 @@ void tileEngineViewMode2D() {
     if (options.orientation == LANDSCAPE) {
         canvas.destRotx = 0.0;
         canvas.destRoty = 0.0;
-        canvas.destRotz = 0.0;
     } else {
         canvas.destRotx = 0.0;
         canvas.destRoty = 0.0;
-        canvas.destRotz = -90.0;
+    }
+    if (canvas.rotation2d == 0) {
+        if (options.orientation == LANDSCAPE) {
+            canvas.destRotz = 0.0;
+        } else {
+            canvas.destRotz = -90.0;
+        }
+        canvas.attractedToRotZ = 1;
     }
 }
 
@@ -579,20 +808,65 @@ void tileEngineViewMode3D() {
     if (options.orientation == LANDSCAPE) {
         canvas.destRotx = LANDSCAPE_ROTATION_X;
         canvas.destRoty = 0.0;
-        canvas.destRotz = 0.0;
     } else {
         canvas.destRotx = 0.0;
         canvas.destRoty = -PORTRAIT_ROTATION_Y;
-        canvas.destRotz = 0.0;
     }
 }
 
+int zoomBarInitialized = FALSE;
 void tileEngineChangeOrientation(Orientation orientation) {
-    options.orientation = orientation;
+    if (orientation != options.orientation) {
+        if (canvas.viewMode == VIEW_2D && canvas.rotation2d == 0) {
+            if (orientation == LANDSCAPE) {
+                canvas.destRotz = 0.0;
+            } else {
+                canvas.destRotz = -90.0;
+            }
+        } else {
+            if (orientation == LANDSCAPE) {
+                canvas.destRotz = canvas.rotz + 90;
+                if (canvas.destRotz > 180) {
+                    canvas.destRotz -= 360;
+                }
+            } else {
+                canvas.destRotz = canvas.rotz - 90;
+                if (canvas.destRotz < -180) {
+                    canvas.destRotz += 360;
+                }
+            }
+        }
+        options.orientation = orientation;
+        if (canvas.viewMode == VIEW_2D) {
+            tileEngineViewMode2D();
+        } else {
+            tileEngineViewMode3D();
+        }
+        canvas.attractedToRotZ = 1;
+        canvas.zoomBarActive = 0;
+    }
+}
+
+void tileEngineToggle2dRotation() {
     if (canvas.viewMode == VIEW_2D) {
-        tileEngineViewMode2D();
+        canvas.rotation2d = 1 - canvas.rotation2d;
+        if (canvas.rotation2d == 0) {
+            canvas.attractedToRotZ = 1;
+            if (options.orientation == LANDSCAPE) {
+                canvas.destRotz = 0.0;
+            } else {
+                canvas.destRotz = -90.0;
+            }
+        } else {
+            canvas.attractedToRotZ = 0;
+        }
     } else {
-        tileEngineViewMode3D();
+        if (options.orientation == LANDSCAPE) {
+            canvas.destRotz = 0.0;
+        } else {
+            canvas.destRotz = -90.0;
+        }
+        canvas.attractedToRotZ = 1;
     }
 }
 
@@ -600,7 +874,7 @@ void shutdownTileEngine() {
     int i;
     for (i = 0; i < DOWNLOADING_THREADS; i++) {
         //        fprintf(stderr, "waiting for download thread %d to finish...\n", SDL_GetThreadID(downloadThreads[i]));
-        SDL_WaitThread(downloadThreads[i], NULL);
+        //SDL_WaitThread(downloadThreads[i], NULL);
     }
 }
 
@@ -691,7 +965,6 @@ void tileEngine_switchTileProvider() {
 
 void tileEngine_computeTilesVisibility() {
     int i, j;
-    int texturesCreateThisFrame = 0;
 
     // tiles share coordinates so it is more efficient to precalculate visibility for each unique coordinate.
     for (i = 0; i < TILES_X + 1; i++) {
@@ -704,15 +977,17 @@ void tileEngine_computeTilesVisibility() {
             if (tileCoordinateVisibility[i][j] || tileCoordinateVisibility[i + 1][j] || tileCoordinateVisibility[i][j + 1]
                     || tileCoordinateVisibility[i + 1][j + 1]) {
                 tiles[i][j][currentTilesIdx] -> visible = TRUE;
-                if (tiles[i][j][currentTilesIdx] -> state == STATE_GL_TEXTURE_NOT_CREATED && texturesCreateThisFrame < 1) {
+                // TODO parametrize how much textures can be created per frame
+                if (tiles[i][j][currentTilesIdx] -> state == STATE_GL_TEXTURE_NOT_CREATED && texturesCreatedThisFrame
+                        < MAX_TEXTURES_CREATED_PER_FRAME) {
                     tiles[i][j][currentTilesIdx]->texture = createTexture(tiles[i][j][currentTilesIdx] -> pixels4444, TILE_SIZE, TILE_SIZE, TRUE);
-                    if (syncLoadedTilesAppearance == 1) {
+                    if (syncLoadedTilesAppearance == TRUE) {
                         tiles[i][j][currentTilesIdx] -> state = STATE_LOADED_SYNC_WAIT;
                     } else {
                         tiles[i][j][currentTilesIdx] -> state = STATE_LOADED;
-                        tiles[i][j][currentTilesIdx] ->stateChangeTime = 0;
+                        //                        tiles[i][j][currentTilesIdx] ->stateChangeTime = 0;
                     }
-                    texturesCreateThisFrame++;
+                    texturesCreatedThisFrame++;
                 }
             } else {
                 tiles[i][j][currentTilesIdx] -> visible = FALSE;
