@@ -21,7 +21,8 @@ void loadTexture(Texture* texture, char* filename) {
     //    fprintf(stderr, "loadTextureFromFile returned: %p\n", textureRGB);
     if (textureRGB != NULL) {
         texture -> size = textureRGB -> w;
-        texture -> name = createTexture(convert8888to4444(textureRGB), texture->size, texture-> size, FALSE);
+        texture -> name = createTexture(convert8888to4444(textureRGB, FALSE), texture->size, texture-> size, FALSE);
+        SDL_FreeSurface(textureRGB);
     } else {
         fprintf(stderr, "loadTexture %s failed\n", filename);
     }
@@ -76,16 +77,32 @@ void drawUiElement(UiElement* elem) {
     glEnable(GL_BLEND);
     // mask
     glBlendFunc(GL_DST_COLOR, GL_ZERO);
-    glColor4f(1.0, 1.0, 1.0, 1.0);
+    glColor4f(1.0, 1.0, 1.0, elem->color.a);
     glBindTexture(GL_TEXTURE_2D, elem->mask.name);
-    setQuadSize(elem->mask.size, elem->mask.size, quadVertices);
+
+    // Dirty hack to make position always the same size on map.
+    if (elem == position) {
+        setQuadSize(elem->mask.size / canvas.scale, elem->mask.size / canvas.scale, quadVertices);
+    } else {
+        setQuadSize(elem->mask.size, elem->mask.size, quadVertices);
+    }
+
     glVertexPointer(3, GL_FLOAT, 0, quadVertices);
     glTexCoordPointer(2, GL_FLOAT, 0, elem->texCoords);
-    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
+
+    if (elem != zoomKnot) {
+        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
+    }
 
     // texture
+
+    // Dirty hack to make the zoom knot transculent
+    if (elem == zoomKnot) {
+        glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
+    } else {
+        glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
+    }
     glColor4f(elem->color.r, elem->color.g, elem->color.b, elem->color.a);
-    glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
     glBindTexture(GL_TEXTURE_2D, elem-> texture.name);
     glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
 
@@ -99,29 +116,39 @@ void loadUI() {
     //    loadTexture(&barbg, "/opt/cloudgps/res/bar-bg.png");
     //    loadTexture(&barbg, "/opt/cloudgps/res/zoom-minus-mask.png");
 
-    zoomOut = createUiElement("/opt/cloudgps/res/zoom-minus.png", "/opt/cloudgps/res/zoom-minus-mask.png", 50, SCREEN_HEIGHT - 50, SCREEN_WIDTH - 50,
-            SCREEN_HEIGHT - 50, &tileEngineZoomOut, &tileEngineZoomOut, TRUE);
+    zoomOut = createUiElement("/opt/cloudgps/res/zoom-minus.png", "/opt/cloudgps/res/zoom-minus-mask.png", 50, SCREEN_HEIGHT - 150, SCREEN_WIDTH
+            - 150, SCREEN_HEIGHT - 50, &tileEngineZoomOut, &tileEngineZoomOut, TRUE);
 
-    zoomIn = createUiElement("/opt/cloudgps/res/zoom-plus.png", "/opt/cloudgps/res/zoom-plus-mask.png", SCREEN_WIDTH - 50, SCREEN_HEIGHT - 50,
-            SCREEN_WIDTH - 50, 50, &tileEngineZoomIn, &tileEngineZoomIn, TRUE);
-    gotomypos = createUiElement("/opt/cloudgps/res/gotomypos.png", "/opt/cloudgps/res/gotomypos-mask.png", SCREEN_WIDTH - 150, SCREEN_HEIGHT - 50,
-            SCREEN_WIDTH - 50, 150, &tileEngineGotomypos, &tileEngineGotomypos, TRUE);
+    zoomIn = createUiElement("/opt/cloudgps/res/zoom-plus.png", "/opt/cloudgps/res/zoom-plus-mask.png", SCREEN_WIDTH - 50, SCREEN_HEIGHT - 150,
+            SCREEN_WIDTH - 150, 50, &tileEngineZoomIn, &tileEngineZoomIn, TRUE);
+    gotomypos = createUiElement("/opt/cloudgps/res/gotomypos.png", "/opt/cloudgps/res/gotomypos-mask.png", SCREEN_WIDTH - 50, SCREEN_HEIGHT - 50,
+            SCREEN_WIDTH - 50, 50, &tileEngineGotomypos, &tileEngineGotomypos, TRUE);
     position = createUiElement("/opt/cloudgps/res/position.png", "/opt/cloudgps/res/position-mask.png", 0, 0, 0, 0, NULL, NULL, TRUE);
-    view2d = createUiElement("/opt/cloudgps/res/2d.png", "/opt/cloudgps/res/2d-mask.png", 150, SCREEN_HEIGHT - 50, SCREEN_WIDTH - 50, SCREEN_HEIGHT
-            - 150, &tileEngineViewMode2D, NULL, TRUE);
+    view2d = createUiElement("/opt/cloudgps/res/2d.png", "/opt/cloudgps/res/2d-mask.png", 50, SCREEN_HEIGHT - 50, SCREEN_WIDTH - 50, SCREEN_HEIGHT
+            - 50, &tileEngineViewMode2D, NULL, TRUE);
     view2d -> status = UI_HIDDEN;
-    view3d = createUiElement("/opt/cloudgps/res/3d.png", "/opt/cloudgps/res/3d-mask.png", 150, SCREEN_HEIGHT - 50, SCREEN_WIDTH - 50, SCREEN_HEIGHT
-            - 150, &tileEngineViewMode3D, NULL, TRUE);
+    view3d = createUiElement("/opt/cloudgps/res/3d.png", "/opt/cloudgps/res/3d-mask.png", 50, SCREEN_HEIGHT - 50, SCREEN_WIDTH - 50, SCREEN_HEIGHT
+            - 50, &tileEngineViewMode3D, NULL, TRUE);
     crosshair = createUiElement("/opt/cloudgps/res/crosshair.png", "/opt/cloudgps/res/crosshair-mask.png", SCREEN_WIDTH / 2, SCREEN_HEIGHT / 2,
             SCREEN_WIDTH / 2, SCREEN_HEIGHT / 2, NULL, NULL, TRUE);
+    compass = createUiElement("/opt/cloudgps/res/compass.png", "/opt/cloudgps/res/compass-mask.png", 150, SCREEN_HEIGHT - 50, SCREEN_WIDTH - 50,
+            SCREEN_HEIGHT - 150, &tileEngineToggle2dRotation, NULL, TRUE);
+    search = createUiElement("/opt/cloudgps/res/search.png", "/opt/cloudgps/res/search-mask.png", SCREEN_WIDTH - 150, SCREEN_HEIGHT - 50,
+            SCREEN_WIDTH - 50, 150, &tileEngineToggleSearchBar, NULL, TRUE);
+    zoomKnot = createUiElement("/opt/cloudgps/res/zoom-knot.png", "/opt/cloudgps/res/zoom-knot-mask.png", SCREEN_WIDTH - 150, SCREEN_HEIGHT - 50,
+            SCREEN_WIDTH - 50, 150, &tileEngineZoomKnotPressed, &tileEngineZoomKnotDragged, TRUE);
+    zoomKnot -> status = UI_HIDDEN;
     volatileUiElems = g_list_append(volatileUiElems, zoomOut);
     volatileUiElems = g_list_append(volatileUiElems, view2d);
     volatileUiElems = g_list_append(volatileUiElems, view3d);
+    volatileUiElems = g_list_append(volatileUiElems, compass);
+    volatileUiElems = g_list_append(volatileUiElems, search);
     volatileUiElems = g_list_append(volatileUiElems, gotomypos);
     volatileUiElems = g_list_append(volatileUiElems, zoomIn);
-
 }
 
+SDL_Surface * texttmpsurface = NULL;
+
 void drawUiElems() {
     GList* elem = uiElems;
 
@@ -132,6 +159,166 @@ void drawUiElems() {
         }
         elem = elem -> next;
     }
+    glTranslatef(compass -> x + compass -> texture.size / 2, compass -> y + compass -> texture.size / 2, 0.0);
+    glDisable(GL_TEXTURE_2D);
+    glEnable(GL_BLEND);
+    glBlendFunc(GL_DST_COLOR, GL_ZERO);
+    if (canvas.viewMode == VIEW_2D && canvas.rotation2d == FALSE) {
+        glColor4f(.2, .2, .2, 0.6);
+    } else {
+        glColor4f(.3, .3, 1.0, 0.6);
+    }
+    glVertexPointer(3, GL_FLOAT, 0, arrowVertices);
+    glRotatef(canvas.rotz, 0.0, 0.0, 1.0);
+    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
+    glRotatef(canvas.rotz, 0.0, 0.0, -1.0);
+
+    glTranslatef(-compass -> x - compass -> texture.size / 2, -compass -> y - compass -> texture.size / 2, 0.0);
+    glEnable(GL_TEXTURE_2D);
+
+    //
+    //    SDL_Color color = { 0, 0, 0 };
+    //    SDL_Surface *surface;
+    //    if (!(surface = TTF_RenderText_Blended(font, "Hello World!", color))) {
+    //        fprintf(stderr, "TTF_RenderText_Shaded failed: %s\n", TTF_GetError());
+    //    } else {
+    //
+    //        if (texttmpsurface == NULL) {
+    //            texttmpsurface = SDL_CreateRGBSurface(SDL_SWSURFACE, 512, 512, 32, 0x000000FF, 0x0000FF00, 0x00FF0000, 0xFF000000);
+    //            if (texttmpsurface == NULL) {
+    //                fprintf(stderr, "createRGBsurface texttmpsurface failed \n");
+    //                exit(1);
+    //            }
+    //        }
+    //        SDL_Rect rect;
+    //        rect.x = 0;
+    //        rect.y = 0;
+    //        rect.w = surface->w;
+    //        rect.h = surface->h;
+    //        SDL_BlitSurface(surface, &rect, texttmpsurface, &rect);
+    //        SDL_FreeSurface(surface);
+    //
+    //
+    //        GLushort * aaa = convert8888to4444(texttmpsurface, FALSE);
+    //        int cropRect[] = {0,0,100, 100};
+    //        GLuint texture = 0;
+    //        glEnable(GL_TEXTURE_2D);
+    //        glGenTextures(1, &texture);
+    //        glBindTexture(GL_TEXTURE_2D, texture);
+    //        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 512, 512, 0, GL_RGBA, GL_UNSIGNED_SHORT_4_4_4_4, aaa);
+    //        glTexParameteriv(GL_TEXTURE_2D, GL_TEXTURE_CROP_RECT_OES, cropRect);//crash
+    //        m_exgldrawte(0, 0, 1, 256, 256);//crash
+    //
+    //
+    //        createTexture(aaa, surface -> w, surface -> h, FALSE);
+    //
+    //        //SDL_BlitSurface(text_surface,NULL,screen,NULL);
+    //        //perhaps we can reuse it, but I assume not for simplicity.
+    //
+    //        SDL_FreeSurface(surface);
+    //    }
+}
+
+void drawZoomBar() {
+    int idx = 0, i;
+
+    if (canvas.zoomBarAlpha == 0.0) {
+        return;
+    }
+    if (canvas.orientationTransitionLinear < 80) {
+        glTranslatef(0, SCREEN_HEIGHT - canvas.searchBarY, 0);
+    } else {
+        glTranslatef(SCREEN_WIDTH - canvas.searchBarY, SCREEN_HEIGHT, 0);
+        glRotatef(-90, 0.0, 0.0, 1.0);
+    }
+    if (canvas.orientationTransitionLinear == 80) {
+        activateZoomBar();
+    }
+
+    GLfloat screenSize = options.orientation == LANDSCAPE ? SCREEN_WIDTH : SCREEN_HEIGHT;
+    int zoomLevelCount = canvas.provider -> maxZoom - canvas.provider -> minZoom + 1;
+
+    if (zoomBarInitialized == FALSE) {
+        for (i = 0; i <= zoomLevelCount; i++) {
+            zoomBarLines[idx++] = i * (screenSize - 200) / zoomLevelCount + 100;
+            zoomBarLines[idx++] = -148;
+            zoomBarLines[idx++] = 0.0;
+
+            zoomBarLines[idx++] = i * (screenSize - 200) / zoomLevelCount + 100;
+            zoomBarLines[idx++] = -158;
+            zoomBarLines[idx++] = 0.0;
+
+        }
+        zoomBarInitialized = TRUE;
+    }
+
+    glDisable(GL_TEXTURE_2D);
+    glEnable(GL_BLEND);
+    glBlendFunc(GL_DST_COLOR, GL_ZERO);
+    glVertexPointer(3, GL_FLOAT, 0, zoomBarLines);
+    glColor4f(1.0 - canvas.zoomBarAlpha, 1.0 - canvas.zoomBarAlpha, 1.0 - canvas.zoomBarAlpha, 1.0);
+    glDrawArrays(GL_LINES, 0, zoomLevelCount * 2 + 2);
+    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+    glColor4f(.4, .4, .4, canvas.zoomBarAlpha);
+    glDrawArrays(GL_LINES, 0, zoomLevelCount * 2 + 2 + 8);
+    glEnable(GL_TEXTURE_2D);
+
+    GLfloat x = canvas.orientationTransitionLinear < 80 ? zoomKnot -> x : SCREEN_HEIGHT - zoomKnot -> y - 64;
+    glTranslatef(x, -180, 0);
+    sprintf(buf, "zoom = %d", getKnotDestZoom());
+    drawStringAlpha(buf, canvas.zoomBarAlpha, canvas.zoomBarAlpha, canvas.zoomBarAlpha, 1.0, FALSE, GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
+    glTranslatef(-x, 180, 0);
+
+    if (canvas.orientationTransitionLinear < 80) {
+        glTranslatef(0, -SCREEN_HEIGHT + canvas.searchBarY, 0);
+    } else {
+        glRotatef(90, 0.0, 0.0, 1.0);
+        glTranslatef(-SCREEN_WIDTH + canvas.searchBarY, -SCREEN_HEIGHT, 0);
+    }
+}
+
+void drawSearchBar() {
+    if (canvas.searchBarY > 0.0) {
+        glDisable(GL_TEXTURE_2D);
+        glEnable(GL_BLEND);
+        if (canvas.orientationTransitionLinear < 80) {
+            glTranslatef(0, SCREEN_HEIGHT - canvas.searchBarY, 0);
+        } else {
+            glTranslatef(SCREEN_WIDTH - canvas.searchBarY, SCREEN_HEIGHT, 0);
+            glRotatef(-90, 0.0, 0.0, 1.0);
+        }
+
+        setQuadSize(canvas.orientationTransitionLinear < 80 ? SCREEN_WIDTH : SCREEN_HEIGHT, 64, quadVertices);
+        glBlendFunc(GL_DST_COLOR, GL_ZERO);
+        glColor4f(.7, .7, .7, 0.5);
+        glVertexPointer(3, GL_FLOAT, 0, quadVertices);
+        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
+
+        glTranslatef(5, 5, 0);
+        setQuadSize(canvas.orientationTransitionLinear < 80 ? SCREEN_WIDTH - 10 : SCREEN_HEIGHT - 10, 64 - 10, quadVertices);
+        glColor4f(0.7, .7, .7, 0.5);
+        glVertexPointer(3, GL_FLOAT, 0, quadVertices);
+        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
+        glTranslatef(5, 25, 0);
+
+        if (((nowMillis - searchBar.lastChange) / 200) % 2 == 0) {
+            GLfloat cursor[] = { searchBar.cursorScreen, -10, 0, searchBar.cursorScreen, 24, 0 };
+            glDisable(GL_BLEND);
+            glColor4f(1.0, 1.0, 1.0, 1.0);
+            glVertexPointer(3, GL_FLOAT, 0, cursor);
+            glDrawArrays(GL_LINES, 0, 2);
+        }
+        glEnable(GL_BLEND);
+        glEnable(GL_TEXTURE_2D);
+        drawString(searchBar.input, 1.0, 1.0, 1.0, FALSE);
+        if (canvas.orientationTransitionLinear < 80) {
+            glTranslatef(-10, -SCREEN_HEIGHT + canvas.searchBarY - 30, 0);
+        } else {
+            glTranslatef(-10, -30, 0);
+            glRotatef(90, 0.0, 0.0, 1.0);
+            glTranslatef(-SCREEN_WIDTH + canvas.searchBarY, -SCREEN_HEIGHT, 0);
+        }
+    }
 }
 
 GLfloat satelliteBars[18 * MAX_SATTELITES];
@@ -408,7 +595,7 @@ void drawStatusBar(Orientation orientation) {
     if (orientation == LANDSCAPE) {
         glTranslatef(-250, -23, 0);
     } else {
-        glTranslatef(125, -30, 0);
+        glTranslatef(135, -30, 0);
     }
 
     glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
@@ -422,7 +609,11 @@ void drawStatusBar(Orientation orientation) {
 
 int processUiMouse() {
     GList* elem = uiElems;
-
+    if (canvas.zoomBarAlpha > 0.2) {
+        zoomKnot -> status = UI_SHOWN;
+    } else {
+        zoomKnot -> status = UI_HIDDEN;
+    }
     if (mouse.button == 1) {
         if (mouse.oldButton == 0) {
             while (elem != NULL) {
@@ -452,6 +643,23 @@ int processUiMouse() {
     return FALSE;
 }
 
+void updateAnimationValue(int enabled, GLfloat *value, GLfloat minValue, GLfloat maxValue, GLfloat speed) {
+    if (enabled) {
+        if (*value < maxValue * (1.0 - speed * speed)) {
+            *value += speed * (maxValue - *value);
+        } else {
+            *value = maxValue;
+        }
+    } else {
+        if (*value > minValue + (maxValue - minValue) * speed * speed) {
+            *value -= speed * (*value - minValue);
+        } else {
+            *value = minValue;
+        }
+    }
+
+}
+
 void updateUi() {
     if (device && device -> fix) {
 
@@ -464,8 +672,8 @@ void updateUi() {
             toTileCoordinate(&wc, &tc, canvas.zoom);
 
             toScreenCoordinate(&tc, &position ->x, &position -> y);
-            position -> x -= position -> texture.size / 2;
-            position -> y -= position -> texture.size / 2;
+            position -> x -= position -> texture.size / 2 / canvas.scale;
+            position -> y -= position -> texture.size / 2 / canvas.scale;
 
             position -> color.a = 0.75 + 0.25 * sin(nowMillis / 500.0);
             position -> color.r = position -> color.a;
@@ -487,27 +695,45 @@ void updateUi() {
     gotomypos -> color.g = canvas.followingMypos * .3 + .4;
     gotomypos -> color.b = canvas.followingMypos * .3 + .4;
 
+    updateAnimationValue(canvas.searchBarActive, &canvas.searchBarY, 0.0, 64.0, 0.1);
+    updateAnimationValue(canvas.zoomBarActive, &canvas.zoomBarAlpha, 0.0, 1.0, 0.03);
+
     GList * volatileElem = volatileUiElems;
     if (canvas.orientationTransitionLinear > 0 && canvas.orientationTransitionLinear < 160) {
         int i = 0;
+
+        if (canvas.orientationTransitionLinear == 80) {
+            zoomBarInitialized = FALSE;
+        }
+
+        if (canvas.searchBarActive) {
+            if (canvas.orientationTransitionLinear <= 80) {
+                canvas.searchBarY = 64 - canvas.orientationTransitionLinear;
+            } else {
+                canvas.searchBarY = 64 - (160 - canvas.orientationTransitionLinear);
+            }
+            if (canvas.searchBarY < 0.0) {
+                canvas.searchBarY = 0.0;
+            }
+        }
         while (volatileElem != NULL) {
             UiElement * elem = volatileElem -> data;
 
-            if (canvas.orientationTransitionLinear < 80) {
+            if (canvas.orientationTransitionLinear <= 80) {
                 elem -> x = elem -> landscapex;
-                elem -> y = elem -> landscapey;
+                elem -> y = elem -> landscapey - canvas.searchBarY;
                 if (canvas.orientationTransitionLinear > i) {
                     GLfloat xxx = canvas.orientationTransitionLinear - i;
-                    elem -> y += pow((xxx), 1.6);
+                    elem -> y += pow((xxx), 1.65);
                 }
                 elem -> texCoords = texCoordsLandscape[0];
             } else {
-                elem -> x = elem -> portraitx;
+                elem -> x = elem -> portraitx - canvas.searchBarY;
                 elem -> y = elem -> portraity;
                 elem -> texCoords = texCoordsPortrait[0];
                 if (160 - canvas.orientationTransitionLinear > i) {
                     GLfloat xxx = 160 - (canvas.orientationTransitionLinear) - i;
-                    elem -> x += pow(xxx, 1.6);
+                    elem -> x += pow(xxx, 1.65);
                 }
             }
             volatileElem = volatileElem -> next;
@@ -518,15 +744,44 @@ void updateUi() {
             UiElement * elem = volatileElem -> data;
             if (options.orientation == LANDSCAPE) {
                 elem -> x = elem -> landscapex;
-                elem -> y = elem -> landscapey;
+                elem -> y = elem -> landscapey - canvas.searchBarY;
                 elem -> texCoords = texCoordsLandscape[0];
             } else {
-                elem -> x = elem -> portraitx;
+                elem -> x = elem -> portraitx - canvas.searchBarY;
                 elem -> y = elem -> portraity;
                 elem -> texCoords = texCoordsPortrait[0];
             }
             volatileElem = volatileElem -> next;
         }
     }
+
+    if (nowMillis > canvas.zoomBarVisibleMilis) {
+        canvas.zoomBarActive = FALSE;
+    }
+
+    if (canvas.orientationTransitionLinear == 79) {
+        activateZoomBar();
+        zoomKnot -> x = canvas.zoomKnotPosition;
+        zoomKnot -> y = SCREEN_HEIGHT - 185 - canvas.searchBarY;
+    } else if (canvas.orientationTransitionLinear == 80) {
+        activateZoomBar();
+        zoomKnot -> y = canvas.zoomKnotPosition;
+        zoomKnot -> x = SCREEN_WIDTH - 185 - canvas.searchBarY;
+    }
+
+    zoomKnot -> color.r = canvas.zoomBarAlpha;
+    zoomKnot -> color.g = canvas.zoomBarAlpha;
+    zoomKnot -> color.b = canvas.zoomBarAlpha;
+    zoomKnot -> color.a = 1.0;
+//    zoomKnot -> status = UI_HIDDEN;
+    if (canvas.orientationTransitionLinear < 80) {
+        zoomKnot -> x += (canvas.zoomKnotPosition - zoomKnot -> x) / 4.0;
+        zoomKnot -> y = SCREEN_HEIGHT - 185 - canvas.searchBarY;
+        zoomKnot -> texCoords =  texCoordsLandscape[0];
+    } else {
+        zoomKnot -> x = SCREEN_WIDTH - 185 - canvas.searchBarY;
+        zoomKnot -> y += (canvas.zoomKnotPosition - zoomKnot -> y) / 4.0;
+        zoomKnot -> texCoords =  texCoordsPortrait[0];
+    }
 }
 
